/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2005-2014 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 * 
 *  上記著作権者は，以下の(1)〜(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 * 
 *  @(#) $Id: target_serial.c 301 2015-01-07 04:57:01Z ertl-ishikawa $
 */

/*
 *  シリアルI/Oデバイス（SIO）ドライバ（BlueNinja用）
 */

#include "kernel/kernel_impl.h"
#include <t_syslog.h>
#include "target_serial.h"
#include "tz1000.h"

/*
 * TODO:
 *  現状ではUART1のみをサポートしている．
 *  今後，Bluetoothの利用などにより
 *  サポートするシリアルポートが増えた
 *  場合の拡張のためにSIOPINIBを残しておく
 */
/*
 *  シリアルI/Oポート初期化ブロックの定義
 */
typedef struct sio_port_initialization_block {
	INTNO    intno;
} SIOPINIB;

/*
 *  シリアルI/Oポート管理ブロックの定義
 */
struct sio_port_control_block {
	const SIOPINIB  *p_siopinib;  /* シリアルI/Oポート初期化ブロック */
	intptr_t        exinf;        /* 拡張情報 */
	bool_t          opnflg;       /* オープン済みフラグ */
};

/*
 * シリアルI/Oポート初期化ブロック
 */
const SIOPINIB siopinib_table[TNUM_SIOP] = {
	{(INTNO)INTNO_SIO}
};

/*
 *  シリアルI/Oポート管理ブロックのエリア
 */
SIOPCB	siopcb_table[TNUM_SIOP];

/*
 *  シリアルI/OポートIDから管理ブロックを取り出すためのマクロ
 */
#define INDEX_SIOP(siopid)	((uint_t)((siopid) - 1))
#define get_siopcb(siopid)	(&(siopcb_table[INDEX_SIOP(siopid)]))

/*
 *  SIOドライバの初期化
 */
void
sio_initialize(intptr_t exinf)
{
	SIOPCB	*p_siopcb;
	uint_t	i;

	/*
	 *  シリアルI/Oポート管理ブロックの初期化
	 */
	for (p_siopcb = siopcb_table, i = 0; i < TNUM_SIOP; p_siopcb++, i++) {
		p_siopcb->p_siopinib = &(siopinib_table[i]);
		p_siopcb->opnflg = false;
	}
}


/*
 *  シリアルI/Oポートのオープン
 */
SIOPCB *
sio_opn_por(ID siopid, intptr_t exinf)
{
	SIOPCB          *p_siopcb;
	bool_t  opnflg;

	p_siopcb = get_siopcb(siopid);

	/*
	 *  オープンしたポートがあるかをopnflgに読んでおく．
	 */
	opnflg = p_siopcb->opnflg;

    if (!opnflg) {
        p_siopcb->exinf = exinf;

        /*
         *  UART1ハードウェアの初期化はtarget_initialize
         *  で一部実施済み
         *  割込み要因マスク(UARTIMSC)は全禁止状態
         */
        /*
         * UART1_UARTIFLSの設定
         *   受信割込み頻度：受信FIFOのx%以上がFULL状態で割込みを起こす
         *     現状1/8に設定
         *   送信割込み頻度：送信FIFOのx%以上がEMPTY状態で割込みを起こす
         *     現状7/8に設定
         */
        sil_wrw_mem(UARTIFLS(UART1_BASE), (UARTIFLS_RX_1 | UARTIFLS_TX_14));
        /* 
         * 割込み要求のクリア
         * UART1_UARTICR の全要因をクリア
         */
        sil_wrw_mem(UARTICR(UART1_BASE), (UARTICR_TXIC | UARTICR_RXIC));

        p_siopcb->opnflg = true;
    }

	return(p_siopcb);
}

/*
 *  シリアルI/Oポートのクローズ
 */
void
sio_cls_por(SIOPCB *p_siopcb)
{
	/*
	 *  シリアルI/O割込みをマスクする．
	 */
	if (!(p_siopcb->opnflg)) {
		dis_int(p_siopcb->p_siopinib->intno);
	}
}

/*
 *  SIOの割込みサービスルーチン
 */
void
sio_isr(intptr_t exinf)
{
	SIOPCB          *p_siopcb;

	p_siopcb = get_siopcb(exinf);

    /* 
     * 割込み要求のクリア
     * UART1_UARTICR の全要因をクリア
     */
    sil_wrw_mem(UARTICR(UART1_BASE), (UARTICR_TXIC | UARTICR_RXIC));

	if ((sil_rew_mem(UARTFR(UART1_BASE)) & UARTFR_RXFE) != UARTFR_RXFE) {
		/*
         *  受信FIFOが空でないならば
		 *  受信通知コールバックルーチンを呼び出す．
		 */
		sio_irdy_rcv(p_siopcb->exinf);
	}
	if ((sil_rew_mem(UARTFR(UART1_BASE)) & UARTFR_TXFF) != UARTFR_TXFF) {
		/*
         *  送信FIFOがFullでないならば
		 *  送信可能コールバックルーチンを呼び出す．
		 */
		sio_irdy_snd(p_siopcb->exinf);
	}
}

/*
 *  シリアルI/Oポートへの文字送信
 */
bool_t
sio_snd_chr(SIOPCB *p_siopcb, char c)
{
	if ((sil_rew_mem(UARTFR(UART1_BASE)) & UARTFR_TXFF) != UARTFR_TXFF) {
		/*
         *  送信FIFOがFullでないならば
		 *  DRに書き込む
		 */
        sil_wrw_mem(UARTDR(UART1_BASE), (uint32_t)((uint8_t)(c)));
		return(true);
	}
    else {
        return(false);
    }
}

/*
 *  シリアルI/Oポートからの文字受信
 */
int_t
sio_rcv_chr(SIOPCB *p_siopcb)
{
    int_t c = -1;
	if ((sil_rew_mem(UARTFR(UART1_BASE)) & UARTFR_RXFE) != UARTFR_RXFE) {
		/*
         *  受信FIFOが空でないならば
		 *  DRを読む
		 */
        c = (int_t)(sil_rew_mem(UARTDR(UART1_BASE)) & 0xFF);
	}
	return(c);
}

/*
 *  シリアルI/Oポートからのコールバックの許可
 */
void
sio_ena_cbr(SIOPCB *p_siopcb, uint_t cbrtn)
{
    switch (cbrtn) {
        case SIO_RDY_RCV:
            /* 受信割込みの許可 */
            sil_orw(UARTIMSC(UART1_BASE), UARTIMSC_RXIM);
            break;
        case SIO_RDY_SND:
            /* 送信割込みの許可 */
            sil_orw(UARTIMSC(UART1_BASE), UARTIMSC_TXIM);
            break;
    }
}

/*
 *  シリアルI/Oポートからのコールバックの禁止
 */
void
sio_dis_cbr(SIOPCB *p_siopcb, uint_t cbrtn)
{
    switch (cbrtn) {
        case SIO_RDY_RCV:
            /* 受信割込みの禁止 */
            sil_andw(UARTIMSC(UART1_BASE), ~UARTIMSC_RXIM);
            break;
        case SIO_RDY_SND:
            /* 送信割込みの禁止 */
            sil_andw(UARTIMSC(UART1_BASE), ~UARTIMSC_TXIM);
            break;
    }
}
